#!/bin/bash
 
# Copyright (c) 2013 Christopher Will<dev@cwill-dev.com>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of 
# this software and associated documentation files (the "Software"), to deal in the 
# Software without restriction, including without limitation the rights to use, 
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
# Software, and to permit persons to whom the Software is furnished to do so, subject 
# to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all 
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 
#####################################################################################
#
# VARIABLES
#
#####################################################################################

# Script name and version
app='MyBack v0.1beta'

# Setting display properties for the desktop notifications
DISPLAY=":0.0"
export DISPLAY

# File defining the backup excludes
MYBACK_BASEDIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )/../" && pwd )/"
EXCLUDE_FILE=${MYBACK_BASEDIR}'excludes.txt'
  
# Will contain all users with a home directory (used for impersonification later on)
declare -a users 

# Virtual Box support
# If set (see --help) then the script will search for all available VirtualBox images 
# of all users. For each image snapshots will get created, so the backup just uses
# incremental changes instead of the complete ie VHD
virtual_box_support=false

# Specify the root directory in which rsync will store the backup (with $date as sub-
# directory). The given storage must be mounted correctly.
#
# As advice, the destination directory should not be located on the same device where 
# the data to backup is located. Because in case of hardware failure you will then 
# not be able to access your backups. Better prefer an external storage.
# Example: /mnt/myBackupDevice/rsync 
dest_dir='' 

# This array will contain all positional arguments (=source directories) specifying 
# each directory that has to get considered during the backup process. 
# In a default and simple environment it should be enough to specify the personal 
# home directory. But it may also list any other directories.
# See the myback_excludes for a default case dealing with a root (/) source directory.
declare -a source_directories

# Will contain the temporary log file (which gets copied to the destination myback
# directory afterwards)
tmp_log_file=''

# We use the current date for each backup as sub-directory-name within the given 
# $dest_dir
date=`date "+%Y-%m-%dT%H:%M:%S"` 
 
 
#####################################################################################
#
# FUNCTIONS
#
#####################################################################################

# Initialize basic variables.
#####################################################################################
function init() {

	# Logfile
	# Redirect stdout & stderr to tmp logfile
	# Gets copied later to the destination directory.
	tmp_rand=$(cat /dev/urandom | tr -cd [:alnum:] | head -c 4)
	tmp_log_file="/tmp/myback_"$date"_"$tmp_rand".log"
	exec >  >(tee -a ${tmp_log_file})
	exec 2> >(tee -a ${tmp_log_file} >&2)
	exec 2>&1

	# Set user list to $users
	# Gets used for snapshotting the virtual box images as well for the
	# UI notification	
	tmp_users="$( getent passwd | grep /home/ | cut -d ':' -f 1 )"
	for u in ${tmp_users}
	do
		tmp_dir="/home/${u}"
		if [ -d "$tmp_dir" ]; then
			users+=(${u})
		fi
	done	
}

# Displays a notification baloon in desktop environment for each user
# Args
# 1 - String - Title
# 2 - String - Content
# 3 - String - System icon (ie dialog-ok, dialog-error, dialog-information etc)
#####################################################################################
function display_notification() {
	icon_dialog="dialog-information"
	if [[ -n "$3" ]]; then
		icon_dialog=${3}
	fi  
	for ((i=0;i<${#users[@]};i++)); do 
		$(su -c "notify-send -u critical \"${1}\" \"${2}\" --icon=${icon_dialog}" -s /bin/sh ${users[i]})
	done  
} 

# Exits script with failure code 1 and displays a desktop notification of type "error" 
# as well.
# Args:
# $1 - String - Error message
#####################################################################################
function failure() {	
	echo $1
	display_notification "MyBack - Error occurred" "${1}" "dialog-error"
	exit 1
}
 
# Prepares the script arguments.
#####################################################################################
function setup_args() { 
	while getopts "h?sd:" opt; do
		case "$opt" in
		
			# help
			h|\?)
				echo $app
				echo 'Simple yet effective rsync backup wrapper'
				echo ''
				echo '[OPTIONS] [DIR1[DIR2[DIRn..]]]'
				echo ''
				echo 'Options:'
				echo '-h Show brief help'
				echo '-d Backups destination directory'
				echo '-s Virtual Box support'
				exit 0
				;;
			
			# destination directory
			d) 
				if [[ -n "$OPTARG" ]]; then
					dest_dir=${OPTARG%/}
				fi 
				;;  
			
			# virtual_box_support
			s) 
				virtual_box_support=true
				;;  
		esac
	done  
	
	# Reset argument index pointer for accessing positional parameters
	shift $(( OPTIND-1 ))
	
	# Iterate through each positional parameter and add it to the source_directories 
	# array
	for var in "$@"
	do
		# Do not remove trailing slash if root directory (/)
		if [ ${var} != '/' ]; then
			var=(${var%/})
		fi
		source_directories+=(${var})
	done
}

# Validates all variables
#####################################################################################
function validate {
 
	# Check whether this script got called as super user. If not, we can not 
	# proceed.
	if [ "$(whoami)" != "root" ]; then
		failure 'Aborting: You must be root to execute this script.'
	fi

	# Check dest_dir to be valid
	if [ ! -d ${dest_dir} ]; then
		failure 'Aborting: Destination directory "'${dest_dir}'" does not exist. Maybe it is not mounted?'
	fi  

	# Check exclude file
	if [ ! -f ${EXCLUDE_FILE} ]; then
		failure 'Aborting: Exclude file "'${EXCLUDE_FILE}'" does not exist.'
	fi 
 
	# Check source directories to be set
	if [ ${#source_directories[@]} == 0 ]; then
		failure 'Aborting: No source directories specified.'
	fi
	
	# Check source directories to be valid
	for ((i=0;i<${#source_directories[@]};i++)); do
		if [ ! -d ${source_directories[i]} ]; then
			failure 'Aborting: Source directory "'${source_directories[i]}'" does not exist. Maybe it is not mounted?'
		fi
	done  
}


#####################################################################################
#
# STARTUP
#
#####################################################################################

# Hello World
echo '========================================'
echo '= THIS SCRIPT COMES WITH NO WARRANTY   ='
echo '= Bash version '${BASH_VERSION}'       ='
echo '= [CTRL] + [Z] to kill process         ='
echo '========================================'

# Do some basic initialization 
init

# Notify UI
process_start_date_time=`date "+%Y-%m-%dT%H:%M:%S"`
display_notification "MyBack - Backup started" "Backup process started at ${process_start_date_time}"

# Handle input parameters
setup_args "$@"
 
# Validate environment
validate 

# Convert source_directories to string (used as rsync parameter)
source_directories_string=''
for ((i=0;i<${#source_directories[@]};i++)); do
	source_directories_string+=${source_directories[i]}' '
done  

# The main symlink pointing to the latest (this) backup
latest_backup_dir=${dest_dir}/current

# The full path to the backup destination directory. Keep hierarchies.
current_bak_dest_dir=${dest_dir}/${date}/

# Do not set the link-dest argument on the first run, since there will be no 
# referencing sub-directories within the "latest backup" directory yet.
link_dest=''
if [ -e ${latest_backup_dir} ]
then 
	link_dest='--link-dest='${latest_backup_dir}/
fi

# Create the directory hierarchy for the current backup
# Create myback special folder
MYBACK_DEST_DIR=${current_bak_dest_dir}_myback_/

# Display working target directory
echo '> Timestamp: '${date}
echo '> Sources: '${source_directories_string}
echo '> Destination: '${dest_dir}/${date}
echo '> Logfile: '${MYBACK_DEST_DIR}myback.log
if ${virtual_box_support}; then
	echo '> VirtualBox support enabled'
else	
	echo '> VirtualBox support disabled'
fi
echo '----------------------------------------'


#####################################################################################
#
# VIRTUALBOX
# Create snapshots of each installed VM, so rsync is able to backup the VirtualBox
# images incrementally
#
#####################################################################################
if ${virtual_box_support}; then
	echo '> Creating VirtualBox snapshots'
	
	# Do snapshots for all users
	for ((i=0;i<${#users[@]};i++)); do
	
		echo "> > User: ${users[i]}"
	
		# Get all installed VMs
		vms=$(su -c "VBoxManage list vms | sed -E 's/^\"(.*)\".*/\1/g'" -s /bin/sh ${users[i]})

		# Create new snapshot of each
		for vm_name in $vms; do 
			echo "> > > Processing ${vm_name}.."
			$(su -c "VBoxManage snapshot ${vm_name} take ${date} --description \"Generated by myback automatically at {$date}\" --pause" -s /bin/sh ${users[i]})
			echo "> > > [DONE] ${vm_name}"
		done
		
	done
	
	echo '> [DONE] Creating VirtualBox snapshots'
fi


#####################################################################################
#
# RSYNC
#
#####################################################################################

# Let the user know which directory we are currently process
echo '> Backing up '${source_directories_string}

# Call rsync - this is the main logic! 
# link_dest             - Is only set if the initial backup was processed already
# current_bak_dest_dir - The backup location (of structure destination/date/hierarchy)
rsync -axPv --exclude-from ${EXCLUDE_FILE} ${link_dest} ${source_directories_string} ${current_bak_dest_dir} 
 
# Keep symlink up to date
if [ -e ${current_bak_dest_dir} ]
then
	rm -f ${latest_backup_dir} # Remove
	ln -s ${dest_dir}/${date} ${latest_backup_dir} # Re-create
fi 


#####################################################################################
#
# RESTORE PREPARATION
# We copy all files that will be required during the recovery to the destination 
# folder as well.
#
#####################################################################################

# Create directories
mkdir -p ${MYBACK_DEST_DIR} 
mkdir ${MYBACK_DEST_DIR}'restore'
mkdir ${MYBACK_DEST_DIR}'restore/apt'
mkdir ${MYBACK_DEST_DIR}'log'
mkdir ${MYBACK_DEST_DIR}'app'
mkdir ${MYBACK_DEST_DIR}'tools'

# Export list of all installed applications and repository keys
apt_dir=${MYBACK_DEST_DIR}'restore/apt/'
cp /etc/apt/sources.list ${apt_dir}'sources.list'
if [ -f '/etc/apt/apt.conf' ]; then
   cp /etc/apt/apt.conf ${apt_dir}'apt.conf'
fi
if [ -f '/etc/apt/preferences' ]; then
	cp /etc/apt/preferences ${apt_dir}'preferences'
fi	
cp -R /etc/apt/sources.list.d/ ${apt_dir}
cp -R /etc/apt/apt.conf.d/ ${apt_dir}
cp -R /etc/apt/preferences.d/ ${apt_dir}
cp -R /var/lib/apt/lists/ ${apt_dir}

apt-key key exportall > ${MYBACK_DEST_DIR}'restore/repositories.keys'

# Solution 1: dpkg selectiom -> Buggy (as of 12.10), re-import won't work
dpkg --get-selections > ${MYBACK_DEST_DIR}'restore/installed-packages.lst' 

# Solution 2: apt-mark -> Buggy as well (as of 12.10)
apt-mark showauto > ${MYBACK_DEST_DIR}'restore/pkgs_auto.lst'
apt-mark showmanual > ${MYBACK_DEST_DIR}'restore/pkgs_manual.lst'

# Solution 3: Manual approach -> Create list of packages to re-install one by on recovery
package_list=$(dpkg-query -Wf '${Package} ')
package_list=${package_list// /$'\n'}  # change the semicolons to white space
for package in $package_list; do
	echo "$package" >> ${MYBACK_DEST_DIR}'restore/apt-install.lst'
done

# Copy tmp logfile to current destination directory
mv ${tmp_log_file} ${MYBACK_DEST_DIR}"log/myback.log"

# Copy myback system files to backup as well
cp -R ${MYBACK_BASEDIR}'app/' ${MYBACK_DEST_DIR}
cp -R ${MYBACK_BASEDIR}'tools/' ${MYBACK_DEST_DIR}
cp -R ${MYBACK_BASEDIR}'restore/' ${MYBACK_DEST_DIR}
cp ${MYBACK_BASEDIR}"excludes.txt" ${MYBACK_DEST_DIR}'/'
cp ${MYBACK_BASEDIR}"exec.sh" ${MYBACK_DEST_DIR}'/'


#####################################################################################
#
# FINISH
#
#####################################################################################

# Notify UI (multiline intended)
process_end_date_time=`date "+%Y-%m-%dT%H:%M:%S"`
finish_msg="Backup process finished at ${process_end_date_time}

Destination directory: ${current_bak_dest_dir}

Logfile: ${MYBACK_DEST_DIR}log/myback.log"
display_notification "MyBack - Backup finished" "${finish_msg}" "dialog-ok"